/**************/
/* XYZ ROUTING */
/**************/
#include "3d_mesh_definitions.h"

/* The XY routing function. Routes from channel c == (x,y,z,p) to processing node d == (dx,dy,z)
   Returns a set of resources.
	As xy routing is deterministic, only one next hop is returned.
   This list of one element is closed with a NULL.
*/
ResourceList* XY_routing(const Resource &c, Procnode* d, const Params* dim, ResourceList* frs) {
    ResourceList *hops = new ResourceListN<2>;
    Resource *nexthops = hops->routed;

	// The coordinates of the destination
	int dx = d->x, dy = d->y, dz = d->z;
	// The coordinates of the processing node at the end of channel c
	int nx = get_end(c).x, ny = get_end(c).y, nz = get_end(c).z;

	if (dx != nx || dy != ny || dz != nz) {				// If we do not arrive at our destination
		if (dx < nx)
			nexthops[0] = Resource(nx,ny,nz,W);	//go west if the destination is west
		else if (dx > nx)
			nexthops[0] = Resource(nx,ny,nz,E); 	//go east if the destination is east
		else if (dy < ny)
			nexthops[0] = Resource(nx,ny,nz,S);	//go south if the destination is southh && in the right column
		else if (dy > ny)
			nexthops[0] = Resource(nx,ny,nz,N);	//go north if the destination is north && in the right column
	}
	return hops;
}


Procnode nearest_tsp(const Resource &c, const Params* dim) {
	Procnode ret;
	if (c.x < dim->x/2) {
		ret.x = 0;
		if (c.y < dim->y/2)
			ret.y = 0;
		else
			ret.y = dim->y-1;
	}
	else {
		ret.x = dim->x-1;
		if (c.y < dim->y/2)
			ret.y = 0;
		else
			ret.y = dim->y-1;
	}
	return ret;
}
/* The routing function. Routes from channel c == (x,y,z,p) to processing node d == (dx,dy,dz)
   Returns a set of resources.
	As routing is deterministic, only one next hop is returned.
   This list of one element is closed with a NULL.
*/
ResourceList* inst_routing(const Resource &c, Procnode* d, const Params* dim, ResourceList* frs) {
    ResourceList *hops = new ResourceListN<2>;
    Resource *nexthops = hops->routed;
	// The coordinates of the destination
	int dz = d->z;
	// The coordinates of the processing node at the end of channel c
	int nx = get_end(c).x, ny = get_end(c).y, nz = get_end(c).z;

	if (nz == dz) {
		// If we are at the right plane: use XY routing
		return XY_routing(c,d,dim,frs);
	}
	else {
		// If we are not at the right plane, find the nearest TSP
		Procnode tsp = nearest_tsp(c, dim);
		if (nx == tsp.x && ny == tsp.y) {
			// If we have arrived at the TSP, route up or down
			if (dz > nz)
				nexthops[0] = Resource(nx,ny,nz,U);	
			else if (dz < nz)
				nexthops[0] = Resource(nx,ny,nz,D);	
		}
		else {
			// Use XY routing to get to the TSP
			return XY_routing(c,&tsp,dim,frs);
		}
	}
	return hops;
}
