/**************/
/* ROUTING    */
/**************/

/*
Routing function for Wireless NoC from UCI.
*/

#include "wireless_vc_faults.c"

int is_border_channel(int x, int y, port p, const Params* dim) {
        return  (x == 0 && (p == W1 || p == W2)) ||
                        (x == dim->x-1 && (p == E1 || p == E2)) ||
                        (y == 0 && (p == N1  || p == N2)) ||
                        (y == dim->y-1 && (p == S1 || p == S2));
}
int is_ok(int x, int y, port p, ResourceList* frs, const Params* dim) {
        return x >=0 && y >= 0 && x < dim->x && y < dim->y && !is_faulty_resource(Resource(x,y,p), frs) && !is_border_channel(x,y,p,dim);
}


// This routing function defines Wireless Routing for a 10x10 NePA architecture.
// We choose a delta:
const int delta = 5;


// Returns true iff the port indicates a wireless channel
bool wireless (const char p) {
	return p == WIRELESS1 || p == WIRELESS2 || p == WIRELESS3 || p == WIRELESS4;
}
// Returns true if [nx,ny] is in the same quadrant as [dx,dy]
bool in_same_quadrant (int nx, int ny, int dx, int dy, const Params* dim) {
	return	(nx < (dim->x /2) && dx < (dim->x /2) && ny < (dim->y /2) && dy < (dim->y /2)) ||
			(nx < (dim->x /2) && dx < (dim->x /2) && ny >= (dim->y /2) && dy >= (dim->y /2)) ||
			(nx >= (dim->x /2) && dx >= (dim->x /2) && ny < (dim->y /2) && dy < (dim->y /2)) ||
			(nx >= (dim->x /2) && dx >= (dim->x /2) && ny >= (dim->y /2) && dy >= (dim->y /2));
}

Resource non_faulty_resource(int x, int y, port p1, port p2, int priority, ResourceList* frs, const Params* dim) {
	switch (priority) {
		case 1:
			if (is_ok(x,y,p1,frs,dim))
				return Resource(x,y,p1);
			else
				return Resource(x,y,p2,S2);
		case 2:
			if (is_ok(x,y,p2,frs,dim))
				return Resource(x,y,p2);
			else
				return Resource(x,y,p1);
	}
	assert(false);
	return Resource(x,y,p1);
}


// The default XY routing function: 
ResourceList* XY_routing (const Resource &c, int dx, int dy, ResourceList* frs, const Params* dim, int subnetwork) {
    ResourceList *hops = new ResourceListN<2>;
    Resource *nexthops = hops->routed;
	int i = 0;

	// The coordinates of the processing node at the end of channel c
	int nx = get_end(c).x, ny = get_end(c).y;

	if (dx > nx)
		nexthops[i++] = non_faulty_resource(nx,ny,E1,E2,subnetwork,frs,dim);
	else if (dx < nx)
		nexthops[i++] = non_faulty_resource(nx,ny,W1,W2,subnetwork,frs,dim);
	else if (dy < ny)
		nexthops[i++] = non_faulty_resource(nx,ny,N1,N2,subnetwork,frs,dim);
	else if (dy > ny)
		nexthops[i++] = non_faulty_resource(nx,ny,S1,S2,subnetwork,frs,dim);

	return hops;
}

// Decide whether the wireless router of the current quadrant is faulty
int wireless_router_of_quadrant_is_not_faulty(int nx, int ny, const Params* dim, ResourceList* frs) {
	// First determine the wireless router in current quadrant
	int wx = nx < (dim->x /2) ? 2 : 7;
	int wy = ny < (dim->y /2) ? 2 : 7;
	return is_ok(wx,wy,WIRELESS1,frs,dim) && is_ok(wx,wy,WIRELESS2,frs,dim) && is_ok(wx,wy,WIRELESS3,frs,dim) && is_ok(wx,wy,WIRELESS4,frs,dim);
}


void compute_distance_to_nearest_wr(int nx, int ny, int dx, int dy, int* wx, int* wy, const Params* dim, ResourceList* frs ) {
	*wx = nx < (dim->x /2) ? 2 : 7;
	*wy = ny < (dim->y /2) ? 2 : 7;
	if (!wireless_router_of_quadrant_is_not_faulty(*wx,*wy,dim,frs)) {
		if (dx >= dim->x/2 && nx < dim->x/2)
			//The packet has to move to another quadrant and may use a wireless router from there
			*wx = 7;
		else if (dx < dim->x/2 && nx >= dim->x/2)
			//The packet has to move to another quadrant and may use a wireless router from there
			*wx = 2;
		else
			// The packet will not use a wireless router
			*wx = 1000000;
	} 
}

// The wireless routing extension:
// Use XY_routing to route towards the wireless router at the center of the quadrant.
// Then transmit the message wirelessly.
ResourceList* WNoC_routing (const Resource &c, int dx, int dy, ResourceList* frs, const Params* dim) {
	// The coordinates of the processing node at the end of channel c
	int nx = get_end(c).x, ny = get_end(c).y;

	// First determine the closest wireless router
	int wx,wy;
	compute_distance_to_nearest_wr(nx,ny,dx,dy,&wx,&wy,dim,frs);

	if (nx != wx || ny != wy)
		// If the packet has not arrived at the wireless router, route it towards it
		return XY_routing(c, wx, wy, frs, dim, 1);
	else {
		// If the packet has arrived at the wireless router, transmit it wirelessly
		// Determine which wireless router:
		port p = dx < (dim->x /2) ? (dy < (dim->y /2) ? WIRELESS4 : WIRELESS3) : (dy < (dim->y /2) ? WIRELESS1 : WIRELESS2);
		// Return a next hop <dx,dy,p>
    	ResourceList *hops = new ResourceListN<2>;
		hops->routed[0] = Resource(nx,ny,p);
		return hops;
	}
}


// Compute the distance from [nx,ny] to [dx,dy] using wires.
int wired_travelling_distance(int nx, int ny, int dx, int dy, const Params* dim) {
	return abs(dx - nx) + abs(dy - ny);
}


// Compute the distance from [nx,ny] to [dx,dy] using wireless transmission.
int wireless_travelling_distance(int nx, int ny, int dx, int dy, const Params* dim, ResourceList* frs) {
		// First determine the closest wireless router
		int wx,wy;
		compute_distance_to_nearest_wr(nx,ny,dx,dy,&wx,&wy,dim,frs);
		// Secondly, determine the receiving wireless router
		int wrecx = dx < (dim->x /2) ? 2 : 7;
		int wrecy = dy < (dim->y /2) ? 2 : 7;
		// Compute the distance to this wireless router
		int distance = abs(wx - nx) + abs(wy - ny);
		// Add one hop for the wireless transmission;
		distance++;
		// Compute the distance from the receiving wireless router to the destination.
		distance += abs(dx - wrecx) + abs(dy - wrecy);
		return distance;
}


// The final routing function:
ResourceList* inst_routing(const Resource &c, Procnode* d, const Params* dim, ResourceList* frs) {
	// The coordinates of the destination
	int dx = d->x, dy = d->y;
	// The coordinates of the processing node at the end of channel c
	int nx = get_end(c).x, ny = get_end(c).y;
	// The distances between source and destination
	int H_B = wired_travelling_distance(nx, ny, dx, dy, dim);
	int H_W = wireless_travelling_distance(nx, ny, dx, dy, dim, frs);

	if (dx != nx || dy != ny) {
		// If the wireless distance is significantly less than the wired distance, choose the wireless route
		// I also need a check that the destination is not in the current quadrant.
		if (H_W < H_B + delta && !in_same_quadrant(nx, ny, dx, dy, dim))
			return WNoC_routing(c, dx, dy, frs, dim);
		else
			return XY_routing(c, dx, dy, frs, dim, 2);
	}
	return new ResourceListN<1>;
}

